"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generate = void 0;
const F = __importStar(require("fp-ts/function"));
const O = __importStar(require("fp-ts/Option"));
const Rx = __importStar(require("rxjs"));
const RxO = __importStar(require("rxjs/operators"));
const Common = __importStar(require("../common"));
const prettier_1 = __importDefault(require("prettier"));
const generate = ({ endpoints$, structures$, enums$, flags$, gateway$, aliases$, }) => F.pipe(Rx.merge(Rx.of(snowflake()), Rx.of(gatewayPayload()), structures$.pipe(
// Ignore GatewayPayload as we have provided our own with generics.
RxO.filter(({ identifier }) => !["GatewayPayload"].includes(identifier)), RxO.map(structure)), 
// Endpoint routes
endpoints$.pipe(RxO.toArray(), RxO.map((routes) => endpoints(routes))), enums$.pipe(RxO.map(enumerable)), flags$.pipe(RxO.map(flags)), gateway$.pipe(RxO.map(gateway)), aliases$.pipe(RxO.map(alias))), RxO.toArray(), RxO.map((chunks) => chunks.join("\n")), RxO.map((source) => prettier_1.default.format(source, { parser: "typescript" })), (ob) => Rx.lastValueFrom(ob));
exports.generate = generate;
const snowflake = () => `export type Snowflake = \`\${bigint}\``;
const gatewayPayload = () => `export interface GatewayPayload<T = any | null> {
  /** opcode for the payload */
  op: GatewayOpcode;
  /** event data */
  d?: T;
  /** sequence number, used for resuming sessions and heartbeats */
  s?: number | null;
  /** the event name for this payload */
  t?: string | null;
}`;
const structure = (s) => {
    const fields = s.fields.map(structureField).join("\n");
    return `export interface ${s.identifier} {
    ${fields}
  }`;
};
const structureField = ({ name, optional, type, description, }) => `/** ${description} */
  "${name}"${optional || type.nullable ? "?" : ""}: ${F.pipe(typeIdentifier(type.identifier), maybeSnowflakeMap(type.snowflakeMap))}${type.array ? "[]" : ""}${type.nullable ? " | null" : ""};`;
const typeIdentifier = (name) => {
    switch (name) {
        case "snowflake":
            return "Snowflake";
        case "integer":
        case "float":
            return "number";
        case "object":
        case "mixed":
        case "array":
            return "any";
        case "timestamp":
            return "string";
        case "dict":
            return "Record<string, string>";
    }
    return name;
};
const maybeSnowflakeMap = (isMap) => (input) => isMap ? `Record<Snowflake, ${input}>` : input;
const enumerable = (e) => `export enum ${e.identifier} {
  ${e.values.map(enumerableValue).join("\n")}
};`;
const enumerableValue = ({ name, value, description }) => {
    const comment = F.pipe(description, O.map((d) => `/** ${d} */\n`), O.getOrElse(() => ""));
    return `${comment}${name} = ${value},`;
};
const flags = (f) => `export const ${f.identifier} = {
  ${f.values.map(flagsValue).join("\n")}
} as const;`;
const flagsValue = ({ name, bigint, left, right, description, }) => {
    const comment = F.pipe(description, O.map((d) => `/** ${d} */\n`), O.getOrElse(() => ""));
    left = bigint ? `BigInt(${left})` : left;
    right = bigint ? `BigInt(${right})` : right;
    return `${comment}${name}: ${left} << ${right},`;
};
const endpoints = (routes) => {
    const objects = routes.map(endpoint);
    const methods = objects.map(({ method }) => method).join("\n");
    const types = objects.map(({ type }) => type).join("\n");
    return `export type Route<P, O> = {
      method: string;
      url: string;
      params?: P;
      options?: O;
    };
    export interface Endpoints<O> {
      ${types}
    }
    export function createRoutes<O = any>(fetch: <R, P>(route: Route<P, O>) => Promise<R>): Endpoints<O> {
      return {
        ${methods}
      };
    }`;
};
const endpoint = ({ route, url, description, method, params, response, }) => {
    const urlParams = F.pipe(O.fromNullable(url.match(/\{.*?\}/g)), O.map((params) => params
        .map((param) => param.replace(/[{}]/, ""))
        .map((param) => Common.typeify(param, false))));
    const urlParamsTyped = F.pipe(urlParams, O.map((params) => params.join(": string, ") + ": string, "), O.getOrElse(() => ""));
    const urlParamsPlain = F.pipe(urlParams, O.map((params) => params.join(", ") + ", "), O.getOrElse(() => ""));
    const urlTemplate = url.replace(/\{(.*?)\}/g, (_, param) => `\${${Common.typeify(param, false)}}`);
    const paramsType = F.pipe(params, O.map(({ identifier, array }) => `Partial<${identifier}${array ? "[]" : ""}>`));
    const paramsArg = F.pipe(paramsType, O.map((type) => `params: ${type}, `), O.getOrElse(() => ""));
    const responseType = F.pipe(response, O.map(({ ref, array }) => {
        const refOrAny = F.pipe(ref, O.map(typeIdentifier), O.getOrElse(() => "any"));
        return `${refOrAny}${array ? "[]" : ""}`;
    }), O.getOrElse(() => "any"));
    const comment = F.pipe(description, O.map((d) => `/** ${d} */\n`), O.getOrElse(() => ""));
    return {
        method: `${route}: (${urlParamsPlain}${paramsArg ? "params, " : ""}options) => fetch({
      method: "${method}",
      url: \`${urlTemplate}\`,${paramsArg ? "\nparams," : ""}
      options,
    }),`,
        type: `${comment}${route}: (${urlParamsTyped}${paramsArg}options?: O) => Promise<${responseType}>;`,
    };
};
const gateway = ({ identifier, values }) => {
    const plural = `${identifier}s`;
    const union = values.map(({ type }) => type).join(" | ");
    const props = values
        .map(({ name, type }) => `"${name}": ${type};`)
        .join("\n");
    return `export type ${identifier} = ${union};
export interface ${plural} {
  ${props}
}`;
};
const alias = ({ identifier, nullable, types, array = false, combinator = "and", }) => {
    const op = combinator === "or" ? "|" : "&";
    const type = types.map(typeIdentifier).join(` ${op} `);
    return `export type ${identifier} = (${type}${nullable ? " | null" : ""})${array ? "[]" : ""};`;
};
